import dataclasses
import json
import logging
import os
import time
from collections.abc import Callable
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import datetime
from types import TracebackType
from typing import Any, Callable, Generator, Iterable, Optional
from tqdm import tqdm


from jf_ingest import logging_helper

logger = logging.getLogger(__name__)

RETRY_EXPONENT_BASE = 5


class RetryLimitExceeded(Exception):
    pass


class StrDefaultEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return str(o)


class IngestIOHelper:
    def __init__(self, s3_bucket: str, s3_path: str, local_file_path: str):
        self.s3_bucket = s3_bucket
        self.s3_path = (s3_path,)
        # EVERYTHING in this file path will (potentially) be uploaded to S3!
        # DO NOT put any creds file in this path!!!!
        self.local_file_path = local_file_path

        if not os.path.exists(self.local_file_path):
            os.makedirs(self.local_file_path)

    def _get_file_name(self, object_name: str, batch_number: Optional[int] = 0):
        return f'{object_name}{batch_number if batch_number else ""}.json'

    def write_json_data_to_local(
        self,
        object_name: str,
        json_data: dict | list[dict],
        batch_number: Optional[int] = 0,
    ):
        try:
            file_name = self._get_file_name(
                object_name=object_name, batch_number=batch_number
            )
            full_file_path = f"{self.local_file_path}/{file_name}"
            logger.debug(f"Attempting to save {object_name} data to {full_file_path}")
            with open(full_file_path, "w") as f:
                f.write(json.dumps(json_data, indent=2, cls=StrDefaultEncoder))
                logger.debug(
                    f"File: {full_file_path}, Size: {round(f.tell() / 1000000, 1)}MB"
                )
        except Exception as e:
            logger.error(
                f"Exception encountered when attempting to write data to local file! Error: {e}"
            )

    def upload_files_to_s3(self):
        # TODO: Write multi threaded uploader to upload this local_file_path
        raise NotImplementedError("Function not implemented!")


def get_wait_time(e: Optional[Exception], retries: int) -> int:
    """
    This function attempts to standardize determination of a wait time on a retryable failure.
    If the exception's response included a Retry-After header, respect it.
    If it does not, we do an exponential backoff - 5s, 25s, 125s.

    A possible future addition would be to add a jitter factor.
    This is a fairly standard practice but not clearly required for our situation.
    """
    # getattr with a default works on _any_ object, even None.
    # We expect that almost always e will be a JIRAError or a RequestException, so we will have a
    # response and it will have headers.
    # So I'm choosing to use the getattr call to handle the valid but infrequent possibility
    # that it may not (None or another Exception type that doesn't have a response), rather tha
    # preemptively checking.
    response = getattr(e, "response", None)
    headers = getattr(response, "headers", {})
    retry_after = headers.get("Retry-After")

    # Normalize retry after if it is a string
    if isinstance(retry_after, str) and retry_after.isnumeric():
        retry_after = int(retry_after)
    # Don't do anything if we have a valid int for retry after
    elif isinstance(retry_after, int):
        pass
    else:
        # Null out any invalid retry after values
        retry_after = None

    if retry_after:
        return retry_after
    else:
        return RETRY_EXPONENT_BASE ** retries


def retry_for_429s(f: Callable[..., Any], *args, max_retries: int = 5, **kwargs) -> Any:
    """
    This function allows for us to retry 429s from Jira. There are much more elegant ways of accomplishing
    this, but this is a quick and reasonable approach to doing so.

    Note:
        - max_retries=5 will give us a maximum wait time of 10m25s.
    """
    for retry in range(max_retries + 1):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            if hasattr(e, "status_code") and e.status_code == 429:
                if retry < max_retries:
                    wait_time = get_wait_time(e, retries=retry)
                    logging_helper.log_standard_error(
                        logging.WARNING,
                        # NOTE: Getting the function name here isn't always useful,
                        # because sometimes we circumvent the JIRA standard library
                        # and use functions like "get" and "_get_json", but it's still
                        # better than nothing
                        msg_args=[f.__name__, retry, max_retries, wait_time],
                        error_code=3071,
                    )
                    time.sleep(wait_time)
                    continue
                else:
                    raise RetryLimitExceeded(e)

            # Raise any non-429 related errors
            raise


def batch_iterable(
    iterable: Iterable, batch_size: int
) -> Generator[list[Any], None, None]:
    """Helper function used for batching a given iterable into equally sized batches

    Args:
        iterable (Iterable): An iterable you want to split into batches
        batch_size (int): The size of the batches you want

    Yields:
        Generator[list[Any], None, None]: This generator yields a list of equal size batches, plus a potential final batch that is less than the batch_size arg
    """
    chunk = []
    i = 0
    for item in iterable:
        chunk.append(item)
        i += 1
        if i == batch_size:
            yield chunk
            chunk = []
            i = 0

    if chunk:
        yield chunk


def format_date_to_jql(datetime: datetime):
    return datetime.strftime("%Y-%m-%d")


class ThreadPoolWithTqdm(ThreadPoolExecutor):
    """THIS CLASS MUST BE USED AS A CONTEXT MANAGER ONLY!!!!
    
    This is a custom class that extends the ThreadPoolExecutor class,
    and combines it with some TQDM (progress bar) functionality. This
    should help reduce the number of repeated code around jf_ingest
    
    Yes, I know there is a concurrency extension for TQDM (tqdm.contrib.concurrent)
    BUT this library only has .map style functions, and it does NOT have a submit style
    function. There are instances where using submit is preferred, and often it lends itself
    to simpler code. That is what this library is for

    Args:
        ThreadPoolExecutor (ThreadPoolExecutor): The parent class that this extends

    Returns:
        ThreadPoolWithTqdm: ThreadPoolWithTqdm
    """

    desc: str
    total: int
    exceptions_encountered: list[Exception]
    prog_bar: tqdm

    def __init__(
        self,
        desc: str = None,
        total: int = 0,
        max_workers: int | None = None,
        thread_name_prefix: str = "",
        initializer: Callable[..., object] | None = None,
        initargs: tuple[Any, ...] = ...,
    ) -> None:
        """Custom Constructor, to allow us to set the progress bar values

        Args:
            desc (str, optional): The description field on the TQDM progress bar. Defaults to None.
            total (int, optional): The total value to set on the TQDM progress bar. Defaults to 0.
            max_workers: The maximum number of threads that can be used to execute the given calls.
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: A callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.
        """
        self.desc = desc
        self.total = total
        self.exceptions_encountered = []
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)

    def __enter__(self):
        """Override the __enter__ function to instantiate a progress bar (TQDM)
        when using this as a context manager.
        """
        self.prog_bar = tqdm(desc=self.desc, total=self.total)
        return super().__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """Override the __exit__ function to tear down a progress bar (TQDM)
        when using this as a context manager.
        Progress bar .close() call MUST BE AFTER THE THREADPOOLEXECUTOR TEARDOWN.
        
        Also raises any exceptions we encountered
        """
        retval = super().__exit__(exc_type, exc_val, exc_tb)

        # Progress bar must be close AFTER the parent __exit__ call,
        # because the parent __exit__ call is blocking
        self.prog_bar.close()

        if self.exceptions_encountered:
            raise self.exceptions_encountered[0]
        return retval

    def submit(self, __fn, *args, **kwargs) -> Future:
        """Submit function override, which allows us to set some custom
        callback functions that handle exceptions and progress bar logic

        Args:
            __fn (_type_): A function to submit to the pool

        Returns:
            Future: A future object returned by super class
        """
        # Get the base class generated future
        future = super().submit(__fn, *args, **kwargs)

        # Add custom callbacks
        future.add_done_callback(self.check_for_exception_callback)
        future.add_done_callback(self.update_progress_bar)
        return future

    def submit_with_custom_callbacks(
        self, __fn, custom_call_backs: list[Callable], *args, **kwargs
    ) -> Future:
        """A custom submit function that allows us to set additional callbacks

        Args:
            __fn (_type_): A function to submit to the pool
            custom_call_backs (list[Callable]): A list of custom callbacks,
            for further customization on this class

        Returns:
            Future: A future object returned by super class
        """
        future = self.submit(__fn, *args, **kwargs)

        for callback in custom_call_backs:
            future.add_done_callback(callback)

        return future

    def check_for_exception_callback(self, future: Future):
        """Helper function for dealing with exceptions. Stores them in a class
        field and raises them on exit
        
        Args:
            future (Future): A ThreadPoolExecutor Future object
        """
        exception = future.exception()
        if exception:
            self.exceptions_encountered.append(exception)

    def update_progress_bar(self, future: Future):
        """A custom callback function that iterates on the progress bar by
        default. Does some guessing to see how much we should updated the 
        progress bar by

        Args:
            future (Future): A ThreadPoolExecutor Future object
        """
        if future.exception():
            return

        result = future.result()

        with tqdm.get_lock():
            if isinstance(result, Iterable):
                self.prog_bar.update(len(result))
            else:
                self.prog_bar.update(1)



import os
import logging
import urllib3
import sys
import yaml
from datetime import datetime

from jf_ingest.jf_jira.auth import JiraAuthConfig
from jf_ingest.jf_jira import JiraIngestionConfig, load_and_push_jira_to_s3


def setup_harness_logging(logging_level: int):
    logging.basicConfig(
        level=logging_level,
        format="%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s"
        if logging_level == logging.DEBUG
        else "%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


# NOTE: This is a work in progress developer debugging tool.
# it is currently run by using the following command:
#   pdm run ingest_harness
# and it requires you to have a creds.env and a config.yml file at
# the root of this project
if __name__ == "__main__":
    debug_mode = "--debug" in sys.argv

    # Get credentials for Auth Config
    auth_config = JiraAuthConfig(
        company_slug=os.getenv("COMPANY_SLUG"),
        url=os.getenv("JIRA_URL"),
        user=os.getenv("JIRA_USERNAME"),
        password=os.getenv("JIRA_PASSWORD"),
    )
    # Get Config data for Ingestion Config
    with open("./config.yml") as yaml_file:
        yaml_data = yaml.safe_load(yaml_file)
        jira_config_data = yaml_data["jira"]
        jira_config_data["earliest_issue_dt"] = (
            datetime.strptime(jira_config_data["earliest_issue_dt"], "%Y-%m-%d")
            if jira_config_data["earliest_issue_dt"]
            else None
        )
        jira_config_data["work_logs_pull_from"] = (
            datetime.strptime(jira_config_data["work_logs_pull_from"], "%Y-%m-%d")
            if jira_config_data["work_logs_pull_from"]
            else datetime.min
        )
        ingest_config = JiraIngestionConfig(**jira_config_data)
        # Inject auth data into config
        ingest_config.auth_config = auth_config
        if ingest_config.earliest_issue_dt == None:
            print(
                "earliest_issue_dt option in config.yml detected as being null, setting to datetime.min!"
            )
            ingest_config.earliest_issue_dt = datetime.min

    setup_harness_logging(logging_level=logging.DEBUG if debug_mode else logging.INFO)

    load_and_push_jira_to_s3(ingest_config)



import logging
import pytest
from unittest.mock import patch
from jira import JIRAError
from jf_ingest.utils import (
    RETRY_EXPONENT_BASE,
    RetryLimitExceeded,
    batch_iterable,
    get_wait_time,
    retry_for_429s,
)

logger = logging.getLogger(__name__)

MOCK_RETRY_AFTER_TIME = 12345


class MockJiraErrorResponseWithRetryAfterAsInt:
    headers = {"Retry-After": MOCK_RETRY_AFTER_TIME}


class MockJiraErrorResponseWithRetryAfterAsStr:
    headers = {"Retry-After": str(MOCK_RETRY_AFTER_TIME)}


class MockJiraErrorResponseWithZeroRetryAfter:
    headers = {"Retry-After": 0}


class MockJiraErrorResponseWithNonIntRetryAfter:
    headers = {"Retry-After": "CAT"}


MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_STR = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsStr(),
)

MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_ZERO = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error with Retry After set to 0",
    response=MockJiraErrorResponseWithZeroRetryAfter(),
)

MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_NOT_INT = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error with Retry After set to a non int",
    response=MockJiraErrorResponseWithNonIntRetryAfter(),
)


def test_get_default_wait_time():
    # Get Default Wait Times
    for retry_num in range(0, 10):
        assert (
            get_wait_time(e=None, retries=retry_num) == RETRY_EXPONENT_BASE ** retry_num
        )


def test_get_wait_times_for_retry_error():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    for retry_num in range(0, 5):
        logger.info(f"Retry number {retry_num}...")
        assert get_wait_time(mock_error, retries=retry_num) == MOCK_RETRY_AFTER_TIME


def test_get_wait_times_for_retry_error_when_retry_is_str():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_STR

    for retry_num in range(0, 5):
        logger.info(f"Retry number {retry_num}...")
        assert get_wait_time(mock_error, retries=retry_num) == MOCK_RETRY_AFTER_TIME


def test_get_wait_times_for_retry_error_non_zero():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_ZERO

    for retry_num in range(0, 10):
        logger.info(f"Retry number {retry_num}...")
        expected_wait_time = RETRY_EXPONENT_BASE ** retry_num
        # Retry times SHOULD NEVER be zero
        assert get_wait_time(mock_error, retries=retry_num) != 0
        assert get_wait_time(mock_error, retries=retry_num) == expected_wait_time


def test_get_wait_times_for_retry_error_non_zero():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_NOT_INT

    for retry_num in range(0, 10):
        logger.info(f"Retry number {retry_num}...")
        expected_wait_time = RETRY_EXPONENT_BASE ** retry_num
        # Retry times SHOULD NEVER be zero
        assert get_wait_time(mock_error, retries=retry_num) != 0
        assert get_wait_time(mock_error, retries=retry_num) == expected_wait_time


def test_get_wait_time_for_non_retry_error():
    mock_non_retry_exception = Exception()
    retries = 5
    assert (
        get_wait_time(mock_non_retry_exception, retries=retries)
        == RETRY_EXPONENT_BASE ** retries
    )


def test_retry_for_429s_timeout():
    def always_raise_429():
        raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        with pytest.raises(RetryLimitExceeded):
            retry_for_429s(always_raise_429)

        try:
            retry_for_429s(always_raise_429)
        except RetryLimitExceeded:
            m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_429s_retry_works():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_429_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_429s(raise_429_three_times) == success_message
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_429s_with_generic_error():
    def raise_non_429():
        raise Exception("Generic Error Testing")

    with pytest.raises(Exception):
        retry_for_429s(raise_non_429)


def test_retry_for_429s_with_jira_non_429_error():
    def raise_non_429():
        raise JIRAError(text="Generic Error Testing", status_code=401)

    with pytest.raises(JIRAError):
        retry_for_429s(raise_non_429)


def test_batch_iterable():
    total_data_size = 85926
    batch_size = 62
    data = [i for i in range(total_data_size)]

    recompiled_batches = []
    for batch in batch_iterable(data, batch_size=batch_size):
        # Assert we are getting proper batch sizes
        assert len(batch) <= batch_size
        recompiled_batches.extend(batch)

    # Assert we got ALL the data
    assert len(recompiled_batches) == total_data_size

    # Assert that data is returned matching and in order
    for i in range(total_data_size):
        assert data[i] == recompiled_batches[i]



import logging

from enum import Enum
from dataclasses import dataclass, field
from datetime import datetime

from jira import Issue

from jf_ingest import diagnostics, logging_helper
from jf_ingest.jf_jira.auth import JiraAuthConfig, JiraAuthMethod, get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    IssueMetadata,
    download_boards_and_sprints,
    download_fields,
    download_issuelinktypes,
    download_issues,
    download_issuetypes,
    download_priorities,
    download_projects_and_versions,
    download_resolutions,
    download_statuses,
    download_users,
    download_worklogs,
)

from jf_ingest.utils import IngestIOHelper, ThreadPoolWithTqdm, batch_iterable

logger = logging.getLogger(__name__)


@dataclass
class JiraIngestionConfig:
    # S3/IO information
    s3_bucket: str
    s3_path: str
    upload_to_s3: bool
    local_file_path: str
    company_slug: str

    # Jira Auth Info
    auth_config: JiraAuthConfig

    # Jira Server Information
    gdpr_active: bool

    # Fields information
    # NOTE: I assumed these are all strs
    include_fields: list[str]
    exclude_fields: list[str]

    # User information
    force_search_users_by_letter: bool
    search_users_by_letter_email_domain: str
    required_email_domains: list[str]
    is_email_required: bool

    # Projects information
    # NOTE: I assumed these are all strs
    include_projects: list[str]
    exclude_projects: list[str]
    include_project_categories: list[str]
    exclude_project_categories: list[str]

    # Boards/Sprints
    download_sprints: bool

    # Issues
    skip_issues: bool
    only_issues: bool
    full_redownload: bool
    earliest_issue_dt: datetime
    issue_download_concurrent_threads: int
    # Dict of Issue ID (str) to IssueMetadata Object
    jellyfish_issue_metadata: list[IssueMetadata]
    jellyfish_project_ids_to_keys: dict

    # worklogs
    download_worklogs: bool
    # Potentially solidify this with the issues date, or pull from
    work_logs_pull_from: datetime

    # Jira Ingest Feature Flags
    feature_flags: dict = field(default_factory=dict)


class JiraObject(Enum):
    JiraFields = "jira_fields"
    JiraProjectsAndVersions = "jira_projects_and_versions"
    JiraUsers = "jira_users"
    JiraResolutions = "jira_resolutions"
    JiraIssueTypes = "jira_issuetypes"
    JiraLinkTypes = "jira_linktypes"
    JiraPriorities = "jira_priorities"
    JiraBoards = "jira_boards"
    JiraSprints = "jira_sprints"
    JiraBoardSprintLinks = "jira_board_sprint_links"
    JiraIssues = "jira_issues"
    JiraIssuesIdsDownloaded = "jira_issue_ids_downloaded"
    JiraIssuesIdsDeleted = "jira_issue_ids_deleted"
    JiraWorklogs = "jira_worklogs"
    JiraStatuses = "jira_statuses"


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_jira_to_s3(config: JiraIngestionConfig):
    if config.skip_issues and config.only_issues:
        logger.warning(
            f"only_issues and skip_issues are both True, so all tasks will be skipped. "
            "Are you sure you know what you're doing?"
        )
        return False

    #######################################################################
    # SET UP JIRA CONNECTIONS (Basic and Potentially Atlassian Connect)
    #######################################################################
    jira_basic_connection = get_jira_connection(
        config=config.auth_config, auth_method=JiraAuthMethod.BasicAuth
    )
    jira_atlas_connect_connection = (
        get_jira_connection(
            config=config.auth_config, auth_method=JiraAuthMethod.AtlassianConnect
        )
        if JiraAuthMethod.AtlassianConnect in config.auth_config.available_auth_methods
        else None
    )
    # There is an ongoing effort to cut all things over to Atlassian Connect only,
    # but it is a piece wise migration for now.
    # OJ-29745
    jira_connect_or_fallback_connection = jira_basic_connection
    if config.feature_flags.get(
        "lusca-auth-always-use-connect-for-atlassian-apis-Q423"
    ):
        jira_connect_or_fallback_connection = jira_atlas_connect_connection

    #######################################################################
    # Init IO Helper
    #######################################################################
    ingest_io_helper = IngestIOHelper(
        s3_bucket=config.s3_bucket,
        s3_path=config.s3_path,
        # TODO: Extract this strftime to be part of the ingest config
        local_file_path=config.local_file_path,
    )

    #######################################################################
    # Jira Projects
    #######################################################################
    projects_and_versions = download_projects_and_versions(
        jira_connection=jira_basic_connection,  # Always use BasicAuth because we want to respect permission restrictions on projects
        jellyfish_project_ids_to_keys=config.jellyfish_project_ids_to_keys,
        jellyfish_issue_metadata=config.jellyfish_issue_metadata,
        include_projects=config.include_projects,
        exclude_projects=config.exclude_projects,
        include_categories=config.include_project_categories,
        exclude_categories=config.exclude_project_categories,
    )

    project_ids = {proj["id"] for proj in projects_and_versions}
    ingest_io_helper.write_json_data_to_local(
        object_name=JiraObject.JiraProjectsAndVersions.value,
        json_data=projects_and_versions,
    )

    if not config.only_issues:
        #######################################################################
        # Jira Fields
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraFields.value,
            json_data=download_fields(
                jira_connect_or_fallback_connection,
                config.include_fields,
                config.exclude_fields,
            ),
        )

        #######################################################################
        # Jira Users
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraUsers.value,
            json_data=download_users(
                jira_basic_connection=jira_basic_connection,  # Use BasicAuth because /users/search is not supported by Connect apps.
                jira_atlas_connect_connection=jira_atlas_connect_connection,  # Use AtlasConnect for 'augment with email' subtask
                gdpr_active=config.gdpr_active,
                search_users_by_letter_email_domain=config.search_users_by_letter_email_domain,
                required_email_domains=config.required_email_domains,
                is_email_required=config.is_email_required,
            ),
        )

        #######################################################################
        # Jira Resolutions
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraResolutions.value,
            json_data=download_resolutions(jira_connect_or_fallback_connection),
        )

        #######################################################################
        # Jira Issue Types
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraIssueTypes.value,
            json_data=download_issuetypes(
                jira_connect_or_fallback_connection, project_ids=project_ids
            ),
        )

        #######################################################################
        # Jira Link Types
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraLinkTypes.value,
            json_data=download_issuelinktypes(jira_connect_or_fallback_connection),
        )

        #######################################################################
        # Jira Priorities
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraPriorities.value,
            json_data=download_priorities(jira_connect_or_fallback_connection),
        )

        #######################################################################
        # Jira Statuses
        #######################################################################
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraStatuses.value,
            json_data=download_statuses(jira_basic_connection),
        )

        #######################################################################
        # Jira Boards, Sprints, and Links
        #######################################################################
        boards, sprints, links = download_boards_and_sprints(
            jira_basic_connection,  # Always use BasicAuth because we want to respect permission restrictions on projects
            config.download_sprints,
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraBoards.value, json_data=boards
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraSprints.value, json_data=sprints
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraBoardSprintLinks.value, json_data=links
        )

    if not config.skip_issues:
        jira_issues, deleted_issue_ids = download_issues(
            jira_connection=jira_connect_or_fallback_connection,
            jira_projects=[proj["key"] for proj in projects_and_versions],
            full_redownload=config.full_redownload,
            earliest_issue_dt=config.earliest_issue_dt,
            issue_download_concurrent_threads=config.issue_download_concurrent_threads,
            jellyfish_issue_metadata=config.jellyfish_issue_metadata,
            include_fields=config.include_fields,
            exclude_fields=config.exclude_fields,
        )

        jira_issues_per_file = 2000
        # Force lazy load to load
        batched_issues: list[list[Issue]] = [
            batch
            for batch in batch_iterable(jira_issues, batch_size=jira_issues_per_file)
        ]
        logger.debug(
            f"Attempting to save {len(batched_issues)} issue batches to local disc..."
        )
        with ThreadPoolWithTqdm(
            desc=f"Saving {len(batched_issues)} batches of jira issues to local disc ({jira_issues_per_file} issues per file)",
            total=len(batched_issues),
            max_workers=5,
            thread_name_prefix="JiraIssuesFileWriter",
        ) as pool:
            for batch_number, issue_batch in enumerate(batched_issues):
                pool.submit(
                    ingest_io_helper.write_json_data_to_local,
                    object_name=JiraObject.JiraIssues.value,
                    json_data=issue_batch,
                    batch_number=batch_number,
                )
        logger.debug(f"Done saving {len(batched_issues)} issue batches to local disc!")

        all_downloaded_issue_ids: list[str] = [issue["id"] for issue in jira_issues]

        # Write Issues
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraIssuesIdsDownloaded.value,
            json_data=[int(issue_id) for issue_id in all_downloaded_issue_ids],
        )

        # Write issues that got deleted
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraIssuesIdsDeleted.value,
            json_data=list(deleted_issue_ids),
        )

        #######################################################################
        # Jira Work Logs
        #######################################################################
        if config.download_worklogs:
            ingest_io_helper.write_json_data_to_local(
                object_name=JiraObject.JiraWorklogs.value,
                json_data=download_worklogs(
                    jira_basic_connection,
                    all_downloaded_issue_ids,
                    config.work_logs_pull_from,
                ),
            )
    else:
        logger.info(
            f"Skipping issues and worklogs bc config.skip_issues is {config.skip_issues}"
        )

    logger.info(f"Data has been saved locally to: {ingest_io_helper.local_file_path}")
    #######################################################################
    # Upload files to S3
    #######################################################################
    if config.upload_to_s3:
        ingest_io_helper.upload_files_to_s3()
    else:
        logger.info(
            f"Not uploading to S3 because upload_to_s3 in the Ingestion config is set to {config.upload_to_s3}"
        )

    return True



import json
import os

import requests_mock

from jf_ingest.jf_jira.auth import JiraAuthConfig, get_jira_connection
from jf_ingest.jf_jira.downloaders import IssueMetadata

_MOCK_REST_BASE_URL = "https://test-co.atlassian.net/rest/api/2"
_MOCK_AGILE_BASE_URL = "https://test-co.atlassian.net/rest/agile/1.0"


def get_jira_mock_connection():

    _MOCK_SERVER_INFO_RESP = (
        '{"baseUrl":"https://test-co.atlassian.net","version":"1001.0.0-SNAPSHOT",'
        '"versionNumbers":[1001,0,0],"deploymentType":"Cloud","buildNumber":100218,'
        '"buildDate":"2023-03-16T08:21:48.000-0400","serverTime":"2023-03-17T16:32:45.255-0400",'
        '"scmInfo":"9999999999999999999999999999999999999999","serverTitle":"JIRA",'
        '"defaultLocale":{"locale":"en_US"}} '
    )

    auth_config = JiraAuthConfig(
        url="https://test-co.atlassian.net/",
        personal_access_token="asdf",
        company_slug="test_co",
    )

    with requests_mock.Mocker() as m:
        _register_jira_uri(m, "serverInfo", f"{_MOCK_SERVER_INFO_RESP}")
        _register_jira_uri_with_file(m, "field", "api_responses/fields.json")
        jira_conn = get_jira_connection(config=auth_config, max_retries=1)

    return jira_conn


def get_fixture_file_data(fixture_path: str):
    with open(f"{os.path.dirname(__file__)}/fixtures/{fixture_path}", "r") as f:
        return f.read()


def get_jellyfish_issue_metadata() -> dict[str, IssueMetadata]:
    with open(
        f"{os.path.dirname(__file__)}/fixtures/jellyfish_issue_metadata.json", "r"
    ) as f:
        return [
            IssueMetadata(**metadata_dict) for metadata_dict in json.loads(f.read())
        ]


def _register_jira_uri_with_file(
    mock: requests_mock.Mocker, endpoint: str, fixture_path: str
):
    _register_jira_uri(mock, endpoint, get_fixture_file_data(fixture_path=fixture_path))


def _register_jira_uri(
    mock: requests_mock.Mocker,
    endpoint: str,
    return_value: str,
    HTTP_ACTION: str = "GET",
    use_agile_endpoint: bool = False,
    status_code: int = 200,
):
    """Helper function used to registering mock results for testing against our mock JIRA API. Works by providing
    a request_mock.Mocker instance, what endpoint you want to spoof, and the string that you want that spoofed endpoint
    to return

    Args:
        mock (requests_mock.Mocker): Mocker object
        endpoint (str): Endpoint to append to the end of the base URL ("https://test-co.atlassian.net/rest/api/2", unless use_agile_endpoint is True)
        return_value (str): The raw str to return from the API endpoint
        use_agile_endpoint (bool, optional): IF TRUE, switch the default base URL from the /rest/api/2 endpoint to the agile endpoint. MUST BE TRUE FOR SPRINT AND BOARD API ENDPOINTS. Defaults to False.
    """

    mock_base_url = (
        _MOCK_REST_BASE_URL if not use_agile_endpoint else _MOCK_AGILE_BASE_URL
    )
    mock.register_uri(
        HTTP_ACTION,
        f"{mock_base_url}/{endpoint}",
        text=return_value,
        status_code=status_code,
    )
